(*+--------------------------------------------------------------------------+
  |                                                                          |
  |                  Calcul d'images de Julia et Mandelbrot                  |
  |                                                                          |
  +--------------------------------------------------------------------------+*)
(* Michel Quercia, le 26/12/98 *)
(* compilation : ocamlopt -o julia unix.cmxa julia.ml -cclib -lunix *)

open Printf;;



               (* +--------------------------------+
                  |  Analyse la ligne de commande  |
                  +--------------------------------+ *)


(* Cherche un fichier dans le PATH *)
let rec path_search nom path_list = match path_list with
  | [] -> ""
  | p::_ when Sys.file_exists (p^"/"^nom) -> p^"/"^nom
  | _::suite -> path_search nom suite
;;

(* dcoupe la chaine PATH en une liste de rpertoires *)
let rec split_path path i =
  if i >= String.length(path) then []
  else begin
    let j = ref(i) in
    while !j < String.length(path) && path.[!j] <> ':' do incr j done;
    String.sub path i (!j-i) :: split_path path (!j+1)
  end
;;
  
(* Cherche un nom dans le rpertoire courant puis dans le PATH *)
let find_file nom =
  if Sys.file_exists nom then nom
  else path_search nom (split_path (Sys.getenv "PATH") 0)
;;



let debug  = ref false
and tmpdir = ref "/tmp"
and jtk    = ref(find_file "jtk")
and fonte  = ref ""
;;

(* analyse la ligne de commande *)
Arg.parse [
  "-d", (Arg.Set debug),                    "dboguage";
  "-f", (Arg.String(fun s -> fonte := s)),  "fonte texte \"famille taille\"";
  "-j", (Arg.String(fun s -> jtk := s)),    "fichier de code Tk";
  "-t", (Arg.String(fun s -> tmpdir := s)), "rpertoire temporaire"
] (fun _ -> ()) "";;

(* dmarre l'interprteur TCL *)
if not(Sys.file_exists !jtk) then failwith "fichier jtk introuvable";;
let (chin, chout) = Unix.open_process "wish";;
let chout_list = if !debug then [chout; stdout] else [chout];;
List.iter (fun ch ->
	     if !fonte <> "" then fprintf ch "set fonte {%s}\n" !fonte;
	     fprintf ch "source %s\n\help_show {}\n" !jtk;
	     flush ch)
  chout_list;;


(* cre un rpertoire temporaire pour stocker les images *)
let base_dir = sprintf "%s/julia-%d/" !tmpdir (Unix.getpid());;
Unix.mkdir base_dir 0o755;;


                  (* +---------------------------+
                     |  Descripteurs des images  |
                     +---------------------------+ *)

type catgorie_image = Julia | Mandelbrot;;
type couleur = int;;

type image = {
    cat : catgorie_image;
    mutable name : string;  (* nom symbolique de la fentre *)
    mutable x0   : float;   (* coordonnes logiques         *)
    mutable y0   : float;   (*    du centre de l'image      *)
    mutable dx   : float;   (* taille logique de l'image    *)
    mutable dy   : float;   (*                              *)
    mutable sx   : float;   (* taille en pixels             *)
    mutable sy   : float;   (*                              *)
    mutable max  : int;	    (* nb maxi d'itrations         *)
    mutable palette : couleur array (* palette de couleurs  *)
  } ;;

type bi_image = { (* images de Mandelbrot et Julia associes *)
    mandelbrot : image;
    julia      : image;
    mutable jx : float;
    mutable jy : float
  } ;;

                       (*+-----------------------------+
                         |  Conversion de coordonnes  |
                         +-----------------------------+*)

let real_from_pixel(image,x,y) =
  ((x/.image.sx -. 0.5)*. image.dx +. image.x0,
   (0.5 -. y/.image.sy) *. image.dy +. image.y0)

and pixel_from_real(image,x,y) =
  (((x -. image.x0)/.image.dx +. 0.5)*.image.sx,
   (0.5 -. (y -. image.y0)/.image.dy)*.image.sy)
;;


                       (*+----------------------------+
                         |  panneau de configuration  |
                         +----------------------------+*)


(* cration *)
let makedialog(image) =
  List.iter (fun ch -> (match image.cat with
  | Julia ->
      fprintf ch
	"makedialog %s {{centre x0 y0} {taille dx dy} {pixels sx sy} {j jx jy} {maxiter max }}\n"
	image.name
  | Mandelbrot ->
      fprintf ch
	"makedialog %s {{centre x0 y0} {taille dx dy} {pixels sx sy} {maxiter max }}\n"
	image.name
  ); flush ch) chout_list
;;

(* actualise les champs *)
let updatedialog(bi_image,image) =
  List.iter (fun ch ->
    fprintf ch
      "setdialog %s {x0 \"%f\" y0 \"%f\" dx \"%f\" dy \"%f\" sx \"%d\" sy \"%d\" jx \"%f\" jy \"%f\" max \"%d\"}\n"
      image.name
      image.x0 image.y0
      image.dx image.dy
      (truncate image.sx) (truncate image.sy)
      bi_image.jx bi_image.jy
      image.max;
    flush ch) chout_list
;;

(* dcode un nombre rel *)
let getval defaut chaine =
  try float_of_string(chaine)
  with _ -> defaut
;;

(* relit les champs *)
let getconfig (bi_image,image) =
  let rec scan = function
  | "x0" :: v :: suite -> image.x0 <- getval image.x0 v; scan suite
  | "y0" :: v :: suite -> image.y0 <- getval image.y0 v; scan suite
  | "dx" :: v :: suite -> image.dx <- getval image.dx v; scan suite
  | "dy" :: v :: suite -> image.dy <- getval image.dy v; scan suite
  | "sx" :: v :: suite -> image.sx <- getval image.sx v; scan suite
  | "sy" :: v :: suite -> image.sy <- getval image.sy v; scan suite
  | "jx" :: v :: suite -> bi_image.jx <- getval bi_image.jx v; scan suite
  | "jy" :: v :: suite -> bi_image.jy <- getval bi_image.jy v; scan suite
  | "max":: v :: suite -> image.max <- truncate(getval (float image.max) v);
                          scan suite
  | _ -> ()
  in scan
;;


                    (*+-----------------------------------+
                      |  Calcul et affichage d'une image  |
                      +-----------------------------------+*)


(* calcule un pixel *)
let pixel(ch,        (* fichier de sortie      *)
          x,y,       (* coordonnes logiques   *)
          jx,jy,     (* paramtres d'itration *)
          nmax,      (* nb maxi d'itrations   *)
	  palette) = (* palette de couleurs    *)

  let u  = ref(x)    and v  = ref(y)    in
  let u2 = ref(x*.x) and v2 = ref(y*.y) in
  let n=ref(0) in

  while !u2 +. !v2 < 4.0 && !n < nmax do
    v  := 2.0*.(!u)*.(!v) +. jy;
    u  := !u2 -. !v2 +. jx;
    u2 := !u*.(!u);
    v2 := !v*.(!v);
    incr(n)
  done;

  let couleur = palette.((!n * (Array.length(palette)-1))/nmax) in
  output_byte ch (couleur lsr 16);
  output_byte ch ((couleur lsr 8) land 255) ;
  output_byte ch (couleur land 255)
;;

(* calcule et affiche une image *)
let draw(bi_image,image) =

  (* change le dessin du pointeur *)
  List.iter (fun ch -> fprintf ch "pointer watch\n"; flush ch) chout_list;


  (* recalcule tous les pixels *)
  let x  = image.x0 -. 0.5*.image.dx
  and y  = image.y0 +. 0.5*.image.dy
  and hx = image.dx /. image.sx
  and hy = image.dy /. image.sy
  and fname = base_dir ^ image.name ^ ".ppm" in
  let ch = open_out_bin fname
  in

  fprintf ch "P6 %d %d 255\n" (truncate image.sx) (truncate image.sy);
  for i=0 to truncate(image.sy)-1 do for j=0 to truncate(image.sx)-1 do
    match image.cat with
    | Mandelbrot ->
	pixel(ch,0.0, 0.0,
                 x +. float(j)*.hx,
                 y -. float(i)*.hy,
                 image.max, image.palette)
    | Julia ->
	pixel(ch,x +. float(j)*.hx,
                 y -. float(i)*.hy,
		 bi_image.jx, bi_image.jy,
                 image.max, image.palette)
  done done;
  close_out ch;

  (* charge l'image *) 
  List.iter (fun ch -> fprintf ch "draw %s %s\n" fname image.name) chout_list;

  (* restaure l'image du pointeur *)
  List.iter (fun ch -> fprintf ch "pointer {}\n"; flush ch) chout_list;

  (* repositionne la croix sur mandelbrot *)
  let (x,y) = pixel_from_real(bi_image.mandelbrot,bi_image.jx,bi_image.jy) in
  List.iter (fun ch ->
    fprintf ch "setcross %s %f %f\n" bi_image.mandelbrot.name x y) chout_list;

  (* actualise l'affichage des champs *)
  updatedialog(bi_image,image)

;;


                 (*+------------------------------+
                   |  Enregistrement, impression  |
                   +------------------------------+*)

let saveimage(image,ext,fname) =

  let source = base_dir ^ image.name ^ ".ppm" in
  let cmd = match ext with
  | "bmp" -> "ppmtobmp"
  | "eps" -> "pnmtops"
  | "gif" -> "ppmtogif"
  | "jpg" -> "cjpeg"
  | "pcx" -> "ppmtopcx"
  | "ppm" -> "cat"
  | "ps"  -> "pnmtops"
  | "tga" -> "ppmtotga"
  | _     -> failwith "format graphique inconnu" in
  let _ = Unix.system(sprintf "%s <%s >%s" cmd source fname) in ()
;;

let printimage(image,commande) =
  let _ = Unix.system(commande ^ " " ^ base_dir ^ image.name ^ ".ppm") in ()
;;



                     (*+----------------------+
                       |  Gestion des images  |
                       +----------------------+*)

(* images initiales *)

let palette = [|
  0xffffff ; 0xfff7ee ; 0xffeedd ; 0xffe6cc ;
  0xffddbb ; 0xffd5aa ; 0xffcc99 ; 0xffc488 ;
  0xffbb77 ; 0xffb366 ; 0xffaa55 ; 0xffa244 ;
  0xff9933 ; 0xff9122 ; 0xff8811 ; 0xff8000 ;
  0xee8811 ; 0xdd9122 ; 0xcc9933 ; 0xbba244 ;
  0xaaaa55 ; 0x99b366 ; 0x88bb77 ; 0x77c488 ;
  0x66cc99 ; 0x55d5aa ; 0x44ddbb ; 0x33e6cc ;
  0x22eedd ; 0x11f7ee ; 0x00ffff ; 0x000000 
|];;


let b0 = {

    mandelbrot = {
        cat  = Mandelbrot;
        name = "mandelbrot";
	x0   = 0.0;
	y0   = 0.0;
	dx   = 3.0;
	dy   = 3.0;
	sx   = 400.0;
	sy   = 400.0;
	max  = 200;
        palette = palette};

    julia = {
        cat  = Julia;
        name = "julia";
	x0   = 0.0;
	y0   = 0.0;
	dx   = 3.0;
	dy   = 3.0;
	sx   = 400.0;
	sy   = 400.0;
	max  = 200;
        palette = palette};

  jx = -0.7998;
  jy = 0.15517
};;


(* liste d'association nom -> image *)
let img_list = ref [("mandelbrot",(b0,b0.mandelbrot));
                    ("julia",     (b0,b0.julia))]
;;

(* duplique une bi-image *)
let new_bi_image(bi_image) =

  let c = string_of_int(List.length(!img_list)/2) in
  let b = {mandelbrot = {bi_image.mandelbrot with name = "mandelbrot-" ^ c};
           julia      = {bi_image.julia      with name = "julia-" ^ c};
           jx         = bi_image.jx;
           jy         = bi_image.jy} in
  img_list := (b.mandelbrot.name,(b,b.mandelbrot)) ::
              (b.julia.name,(b,b.julia))           :: !img_list;
  b
;;


                   (* +------------------------+
                      |  Gestion des palettes  |
                      +------------------------+ *)

let showpalette(image) =

  List.iter (fun ch ->
    fprintf ch "palette %s %d { " image.name (Array.length image.palette);
    for i=0 to Array.length(image.palette) - 1 do
      fprintf ch "#%06x " image.palette.(i)
    done;
    fprintf ch "}\n";
    flush ch)
    chout_list
;;

let getpalette(liste) =

  let n = int_of_string(List.hd liste) in
  let p = Array.make n 0               in
  let l = ref(List.tl liste)           in
  for i=0 to n-1 do
    p.(i) <- int_of_string("0x" ^ String.sub (List.hd !l) 1 6);
    l := List.tl !l
  done;
  p
;;



                  (*+-----------------------------+
                    |  Interprteur de commandes  |
                    +-----------------------------+*)

(* dcoupe une chaine en mots *)
let rec dcoupe(s,i) =
  if i >= String.length(s) then []
  else match s.[i] with
  | (' '|'\t'|'\n') -> dcoupe(s,i+1)
  | '"'  -> chaine(s,i+1,i+1)
  | _ -> mot(s,i,i+1)
and mot(s,i,j) =
  if j >= String.length(s) then [String.sub s i (j-i)]
  else match s.[j] with
  | (' '|'\t'|'\n') -> String.sub s i (j-i) :: dcoupe(s,j+1)
  | _ -> mot(s,i,j+1)
and chaine(s,i,j) =
  if j >= String.length(s) then [String.sub s i (j-i)]
  else match s.[j] with
  | '"' -> String.sub s i (j-i) :: dcoupe(s,j+1)
  | _   -> chaine(s,i,j+1)
;;

      
let f = float_of_string;;

(* xcute une commande sur une image *)
let rec commande (bi_image,image) params = match params with

| ["resize"; w; h] ->
    let w = f(w)
    and h = f(h) in
    if w <> image.sx or h <> image.sy then begin
      image.dx <- image.dx *. w /. image.sx;
      image.dy <- image.dy *. h /. image.sy;
      image.sx <- w;
      image.sy <- h;
      draw (bi_image,image)
    end

| ["zoom"; x0; y0; r] ->
    let r = f(r)
    and (x,y) = real_from_pixel(image,f(x0),f(y0)) in
    image.x0 <- x;
    image.y0 <- y;
    image.dx <- r *. image.dx;
    image.dy <- r *. image.dy;
    draw (bi_image,image)

| ["move"; rx; ry] ->
    image.x0 <- image.x0 +. f(rx)*.image.dx;
    image.y0 <- image.y0 +. f(ry)*.image.dy;
    draw (bi_image,image)

| ["point"; x; y] when image.cat = Mandelbrot ->
    let (x,y) = real_from_pixel(image,f(x),f(y)) in
    bi_image.jx <- x;
    bi_image.jy <- y;
    draw (bi_image,bi_image.julia)

| ["showconfig"] ->
    makedialog(image);
    updatedialog(bi_image,image)

| ["showmandelbrot"] ->
    makedialog(bi_image.mandelbrot);
    updatedialog(bi_image,bi_image.mandelbrot)

| ["showjulia"] ->
    makedialog(bi_image.julia);
    updatedialog(bi_image,bi_image.julia)

| ["showpalette"] ->
    showpalette(image)

| "configure" :: suite ->
    getconfig (bi_image,image) suite;
    draw (bi_image,image)

| "palette" :: suite ->
    image.palette <- getpalette(suite);
    draw(bi_image,image)

| "new" :: suite ->
    updatedialog(bi_image,bi_image.mandelbrot);
    updatedialog(bi_image,bi_image.julia);
    let b = new_bi_image(bi_image) in
    let i = match image.cat with Julia -> b.julia | _ -> b.mandelbrot in
    commande (b,i) suite

| ["destroy"] ->
    Unix.unlink (base_dir ^ image.name ^ ".ppm")

| ["save"; ext; fname] ->
    saveimage(image,ext,fname)

| ["print"; commande] ->
    printimage(image,commande)

|  _ -> ()
;;


(* boucle d'interaction *)
let boucle() =

  let encore = ref true in
  while !encore do
    let ligne = input_line chin in
    if !debug then (printf ">%s\n" ligne; flush stdout);
    match dcoupe(ligne,0) with
    | ["exit"] -> encore := false
    | key::suite -> commande (List.assoc key !img_list) suite
    | _ -> failwith "commande imprvue"
  done

;;

(* on y va ... *)
makedialog(b0.mandelbrot);;
updatedialog(b0,b0.mandelbrot);;
makedialog(b0.julia);;
updatedialog(b0,b0.julia);;
boucle();;
List.iter (fun ch -> fprintf ch "exit\n"; flush ch) chout_list;;
Unix.execvp "rm" [|"rm"; "-R"; base_dir|];;

